// drweb-clamd API usage example.
// Use separate ClamD connection for each file.
// (C) 2003-2016 Doctor Web. All rights reserved.

#include <algorithm>
#include <cstring>
#include <iostream>
#include <signal.h>
#include <string>

#include <assert.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "drweb_clamd_api_common.hpp"

static void send_file(int clamd, int fd)
{
    char buf[4096];
    while(true) {
        ssize_t size = read(fd, buf, sizeof(buf) / sizeof(*buf));
        if(size == -1) {
            if(errno == EINTR) continue;
            THROW("ERROR: read: " << strerror(errno));
        }
        else if(size == 0) break;
        write_exact(clamd, buf, size);
    }
}

static void stream_scan(int clamd, const std::string& path)
{
    std::string cmd = "nSTREAM\n";
    write_exact(clamd, cmd.c_str(), cmd.size());
    in_port_t port;
    {
        std::istringstream is(read_until(clamd, '\n'));
        std::string port_str;
        is >> port_str >> port;
        if(is.fail() || port_str != "PORT")
            THROW("ERROR: STREAM: protocol violation by ClamD");
    }

    scoped_fd_t file = open(path.c_str(), O_RDONLY);
    if(file == -1) THROW("ERROR: open(" << path << "): " << strerror(errno));

    scoped_fd_t sock = socket(AF_INET, SOCK_STREAM, 0);
    if(sock == -1) THROW("ERROR: socket: " << strerror(errno));

    sockaddr_in addr;
    std::memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = htonl((127 << 24) | 1);
    if(connect(sock, reinterpret_cast<sockaddr*>(&addr), sizeof(addr)) == -1)
        THROW("ERROR: connect(127.0.0.1:" << port << "): " << strerror(errno));

    send_file(sock, file);
}

int main(int argc, const char* argv[])
{
    try {
        signal(SIGPIPE, SIG_IGN);

        const std::vector<std::string> supported_cmds =
            { "SCAN", "CONTSCAN", "MULTISCAN", "INSTREAM", "FILDES", "STREAM" };
        parsed_args_t args = parse_args(argc, argv, supported_cmds);

        for(const auto& file : args.files) {
            scoped_fd_t clamd = connect_to_clamd(args.socket_path.c_str());

            if(args.command == "SCAN" || args.command == "CONTSCAN" || args.command == "MULTISCAN")
                scan(clamd, file, args.command);
            else if(args.command == "INSTREAM")
                instream_scan(clamd, file);
            else if(args.command == "FILDES")
                fildes_scan(clamd, file);
            else if(args.command == "STREAM")
                stream_scan(clamd, file);
            else assert(false);

            read_response(clamd);
        }
        return 0;
    } catch(const std::exception& ex) {
        std::cerr << ex.what() << std::endl;
        return 1;
    }
}
