// drweb-clamd API usage example.
// (C) 2003-2016 Doctor Web. All rights reserved.

#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/un.h>
#include <unistd.h>

#include "drweb_clamd_c_api_common.h"

extern const char* supported_cmds[];

void error(const char* fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    vfprintf(stderr, fmt, args);
    fputc('\n', stderr);
    va_end(args);
    _exit(1);
}

int connect_to_clamd(const char* socket_path)
{
    int sock = socket(AF_LOCAL, SOCK_STREAM, 0);
    if(sock == -1) error("ERROR: socket: %s", strerror(errno));

    struct sockaddr_un addr = {0};
    addr.sun_family = AF_LOCAL;
    size_t max_len = sizeof(addr.sun_path) / sizeof(*addr.sun_path) - 1;
    if(strlen(socket_path) > max_len)
        error("ERROR: socket path is too long: %s", socket_path);
    strcpy(addr.sun_path, socket_path);

    if(connect(sock, (struct sockaddr*)(&addr), sizeof(addr)) == -1)
        error("ERROR: connect(%s): %s", socket_path, strerror(errno));

    return sock;
}

void write_exact(int fd, const void* buffer, size_t size)
{
    ssize_t total = 0;
    do {
        ssize_t part = write(fd, buffer + total, size - total);
        if(part > 0)
            total += part;
        else if(part == 0 && size > 0)
            error("ERROR: write: unexpected error");
        else if(part == -1) {
            if(errno == EINTR) continue;
            error("ERROR: write: %s", strerror(errno));
        }
    } while(total > 0 && (size_t)total < size);
}

void scan(int clamd, const char* file_path, const char* cmd)
{
    char scan_cmd[1 + strlen(cmd) + 1 + strlen(file_path) + 2];
    sprintf(scan_cmd, "n%s %s\n", cmd, file_path);
    write_exact(clamd, scan_cmd, strlen(scan_cmd));
}

static void send_by_chunks(int clamd, int fd)
{
    char buf[4096];
    while(1) {
        ssize_t size = read(fd, buf + 4, sizeof(buf) / sizeof(*buf) - 4);
        if(size == -1) {
            if(errno == EINTR) continue;
            error("ERROR: read: %s", strerror(errno));
        } else if(size == 0) {
            write_exact(clamd, "\x00\x00\x00\x00", 4);
            break;
        }
        uint32_t* header = (uint32_t*)buf;
        *header = htonl(size);
        write_exact(clamd, buf, size + 4);
    }
}

void instream_scan(int clamd, const char* file_path)
{
    const char* cmd = "nINSTREAM\n";
    write_exact(clamd, cmd, strlen(cmd));

    int file = open(file_path, O_RDONLY);
    if(file == -1) error("ERROR: open(%s): %s", file_path, strerror(errno));

    send_by_chunks(clamd, file);
}

static void send_fd(int clamd, int fd)
{
    struct iovec iov;
    char garbage = '\x00';
    iov.iov_base = &garbage;
    iov.iov_len = 1;

    char cmsgbuf[CMSG_SPACE(sizeof(fd))];

    struct msghdr mh = {0};
    mh.msg_iov = &iov;
    mh.msg_iovlen = 1;
    mh.msg_control = cmsgbuf;
    mh.msg_controllen = sizeof(cmsgbuf);

    struct cmsghdr* cmsg = CMSG_FIRSTHDR(&mh);
    cmsg->cmsg_level = SOL_SOCKET;
    cmsg->cmsg_type = SCM_RIGHTS;
    cmsg->cmsg_len = CMSG_LEN(sizeof(int));
    int* payload = (int*)CMSG_DATA(cmsg);
    *payload = fd;

    mh.msg_controllen = cmsg->cmsg_len;

    int result = sendmsg(clamd, &mh, 0);
    if(result == -1)
        error("ERROR: sendmsg: %s", strerror(errno));
    else if(result == 0)
        error("ERROR: sendmsg: unexpected error");
}

void fildes_scan(int clamd, const char* file_path)
{
    const char* cmd = "nFILDES\n";
    write_exact(clamd, cmd, strlen(cmd));

    int file = open(file_path, O_RDONLY);
    if(file == -1) error("ERROR: open(%s): ", file_path, strerror(errno));

    send_fd(clamd, file);
}

static void usage(const char* exe)
{
    int buf_size = 0;
    const char* separator = " | ";
    for(int i = 0; supported_cmds[i] != NULL; ++i)
        buf_size += strlen(supported_cmds[i]) + 1 + strlen(separator);

    char buf[buf_size];
    int offset = 0;
    for(int i = 0; supported_cmds[i] != NULL; ++i) {
        strcpy(buf + offset, supported_cmds[i]);
        offset += strlen(supported_cmds[i]);
        strcpy(buf + offset, separator);
        offset += strlen(separator);
    }
    buf[offset - strlen(separator)] = '\0';

    error("Usage: %s <socket path> <files...> [-c %s]", exe, buf);
}

void parse_args(int argc, const char* argv[], parsed_args_t* args)
{
    if(argc < 3) usage(argv[0]);

    args->socket = argv[1];
    args->files[0] = argv[2];
    args->cmd = NULL;
    int file = 1;
    for(int arg = 3; arg < argc; ++arg) {
        if(strcmp(argv[arg], "-c") == 0) {
            if(++arg == argc) usage(argv[0]);
            for(int i = 0; supported_cmds[i] != NULL; ++i)
                if(strcmp(argv[arg], supported_cmds[i]) == 0)
                    args->cmd = argv[arg];
            if(!args->cmd) usage(argv[0]);
        } else {
            if(file == MAX_FILES) error("ERROR: file limit exceeded: %d", MAX_FILES);
            args->files[file++] = argv[arg];
        }
    }
    args->files[file] = NULL;
    if(!args->cmd) args->cmd = supported_cmds[0];
}

void read_response(int fd)
{
    char buf[4096];
    while(1) {
        ssize_t size = read(fd, buf, sizeof(buf) / sizeof(*buf));
        if(size == -1) {
            if(errno == EINTR) continue;
            error("ERROR: read: %s", strerror(errno));
        }
        else if(size == 0) break;
        fwrite(buf, sizeof(char), size, stdout);
    }
}

size_t read_until(int sock, char term, char* buf, size_t size)
{
    size_t total_read = 0;
    while(1) {
        if(size - total_read <= 0)
            error("ERROR: buffer size %d exceeded while trying to read until '%d'", size, (int)term);
        ssize_t just_read = recv(sock, buf + total_read, size - total_read, MSG_PEEK);
        if(just_read == -1) {
            if(errno == EINTR) continue;
            error("ERROR: recv: %s", strerror(errno));
        }
        else if(just_read == 0) break;
        total_read += just_read;

        char* pos = memchr(buf, term, just_read);
        int found = pos != NULL;
        ssize_t consume = found ? pos + 1 - buf : just_read;
        recv(sock, buf, consume, 0);
        if(found) break;
    }
    return total_read;
}
